# -*- coding: utf-8 -*-
"""
Logging utilities
"""
import sys
import logging
import logging.handlers
from tempfile import gettempdir

from pathlib import Path
from PyQt5 import QtCore

from .. import __version__

FORMATTER = logging.Formatter("%(asctime)s | %(name)s | %(levelname)s | %(message)s")

LOG_DIRECTORY = Path(gettempdir()) / f"iris-{__version__}-logs"
LOG_DIRECTORY.mkdir(exist_ok=True)


class QLogger(QtCore.QObject):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.logger = logging.getLogger()
        self.logger.setLevel(logging.DEBUG)

        self.file_handler = logging.handlers.TimedRotatingFileHandler(
            filename=LOG_DIRECTORY / "iris.log", when="midnight", backupCount=5
        )
        self.file_handler.setLevel(logging.DEBUG)
        self.file_handler.setFormatter(FORMATTER)

        self.stream_handler = logging.StreamHandler(stream=sys.stdout)
        self.stream_handler.setFormatter(logging.INFO)
        self.stream_handler.setFormatter(FORMATTER)

        self.logger.addHandler(self.file_handler)
        self.logger.addHandler(self.stream_handler)

    def log(self, msg, level):
        """ General method for logging. Ideal for dynamic logging levels. """
        self.logger.log(level, msg)

    # The methods below mirror those from the logging module
    @QtCore.pyqtSlot(str)
    def debug(self, msg):
        return self.logger.debug(msg)

    @QtCore.pyqtSlot(str)
    def info(self, msg):
        return self.logger.info(msg)

    @QtCore.pyqtSlot(str)
    def warning(self, msg):
        return self.logger.warning(msg)

    @QtCore.pyqtSlot(str)
    def error(self, msg):
        return self.logger.error(msg)

    @QtCore.pyqtSlot(str)
    def critical(self, msg):
        return self.logger.critical(msg)

    @QtCore.pyqtSlot(str)
    def exception(self, msg):
        return self.logger.exception(msg)



# -*- coding: utf-8 -*-
import numpy as np

from . import TestRawDataset
from iris import AbstractRawDataset
from iris.meta import ExperimentalParameter
import pytest


def test_raw_abstract_methods():
    """Test that instantiation of AbstractRawDataset
    raises an error"""
    with pytest.raises(TypeError):
        AbstractRawDataset("")


def test_raw_minimal_methods():
    """
    Test implementing the minimal methods:

    * raw_data
    """
    TestRawDataset()


def test_raw_data_bounds():
    """ Test that a ValueError is raised if ``timedelay`` or ``scan`` are out-of-bounds. """
    test_dataset = TestRawDataset()

    with pytest.raises(ValueError):
        test_dataset.raw_data(timedelay=20, scan=1)

    with pytest.raises(ValueError):
        test_dataset.raw_data(timedelay=5, scan=-1)


def test_raw_experimental_parameters():
    """ Test the behavior of the ExperimentalParameter descriptor """

    test_dataset = TestRawDataset()

    assert test_dataset.test == 0

    test_dataset.test = 1
    assert test_dataset.test == 1

    with pytest.raises(TypeError):
        test_dataset.test = "test"


def test_raw_valid_metadata():
    """ Test that the class attribute 'valid_metadata' is working as intended """

    assert "test" in TestRawDataset.valid_metadata
    assert AbstractRawDataset.valid_metadata <= TestRawDataset.valid_metadata


def test_raw_init_metadata():
    """Test that metadata is recorded correctly inside __init__ and
    that invalid metadata is ignored."""
    test_dataset = TestRawDataset(
        metadata={"test": 5, "fluence": -2, "random_attr": None}
    )
    assert test_dataset.test == 5
    assert test_dataset.fluence == -2

    # Invalid metadata should be ignored.
    assert not hasattr(test_dataset, "random_attr")



# -*- coding: utf-8 -*-
"""
Dialog for processing between AbstractRawDataset and DiffractionDataset
"""
from os import cpu_count

import numpy as np
import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets

fletcher32_help = """ Adds a checksum to each chunk to detect data corruption. 
Attempts to read corrupted chunks will fail with an error. 
No significant speed penalty """.replace(
    "\n", ""
)

shuffle_help = """ Block-oriented compressors like GZIP or LZF work better 
when presented with runs of similar values. Enabling the 
shuffle filter rearranges the bytes in the chunk and may 
improve compression ratio. """.replace(
    "\n", ""
)

alignment_help = """If checked, diffraction patterns will be aligned 
using masked normalized cross-correlation. 
This can double the processing time. """.replace(
    "\n", ""
)

normalization_help = """If checked, diffraction patterns are normalized so that the total
 intensity is equal for each picture at the same scan. For this to be effective, a good mask 
must be provided. """.replace(
    "\n", ""
)

exclude_scans_help = """ Specify scans to exclude comma separated,
e.g. 3,4, 5, 10, 32. """.replace(
    "\n", ""
)

DTYPE_NAMES = {
    "Auto": None,
    "64-bit floats": np.float64,
    "32-bit floats": np.float32,
    "16-bit floats": np.float16,
    "64-bit integers": np.int64,
    "32-bit integers": np.int32,
    "16-bit integers": np.int16,
}


class MaskCreator(QtWidgets.QWidget):
    """ Widget allowing for creation of arbitrary masks """

    def __init__(self, image, **kwargs):
        super().__init__(**kwargs)

        self.resolution = np.array(image.shape)

        self.rect_masks = list()  # list of pg.ROI objects
        self.circ_masks = list()

        self.viewer = pg.ImageView(parent=self)
        self.viewer.setImage(image)
        self.viewer.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding,
            QtWidgets.QSizePolicy.MinimumExpanding,
        )

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.viewer)
        self.setLayout(layout)

    @QtCore.pyqtSlot()
    def add_rect_mask(self):
        new_roi = pg.RectROI(
            pos=self.resolution / 2,
            size=self.resolution / 10,
            pen=pg.mkPen("r", width=4),
        )
        new_roi.addScaleHandle([1, 1], [0, 0])
        new_roi.addScaleHandle([0, 0], [1, 1])
        self.viewer.addItem(new_roi)
        self.rect_masks.append(new_roi)

    @QtCore.pyqtSlot()
    def add_circ_mask(self):
        new_roi = pg.CircleROI(
            pos=self.resolution / 2,
            size=self.resolution / 10,
            pen=pg.mkPen("r", width=4),
        )
        self.viewer.addItem(new_roi)
        self.circ_masks.append(new_roi)

    @QtCore.pyqtSlot()
    def show_preview_mask(self):
        image = np.array(self.viewer.image)
        mask = self.composite_mask()
        image[mask] = 0.0

        dialog = QtWidgets.QDialog(parent=self)
        dialog.setWindowTitle("Mask preview")
        dialog.setModal(True)

        view_widget = pg.ImageView(parent=dialog)
        view_widget.setImage(image)

        return dialog.exec_()

    @QtCore.pyqtSlot()
    def clear_masks(self):
        for roi in self.rect_masks + self.circ_masks:
            self.viewer.removeItem(roi)
        self.rect_masks.clear()
        self.circ_masks.clear()

    def composite_mask(self):
        """ Returns composite mask where invalid pixels are marked as True """
        # Initially, all pixels are valid
        mask = np.zeros(self.resolution, dtype=bool)

        if len(self.circ_masks + self.rect_masks) == 0:
            return mask

        # No need to compute xx, yy, and rr if there are no
        # circular masks, hence the check for empty list
        if self.circ_masks:
            xx, yy = np.meshgrid(
                np.arange(0, mask.shape[0]), np.arange(0, mask.shape[1])
            )
            rr = np.empty_like(xx)

            # Calculating the center position assumes that PyQtGraph is configured
            # such that imageAxisOrder == 'row-major'
            for circ_mask in self.circ_masks:
                radius = circ_mask.size().x() / 2
                corner_x, corner_y = circ_mask.pos().x(), circ_mask.pos().y()
                xc, yc = (round(corner_x + radius), round(corner_y + radius))
                rr = np.hypot(xx - xc, yy - yc)
                mask[rr <= radius] = 1

        for rect_mask in self.rect_masks:
            (x_slice, y_slice), _ = rect_mask.getArraySlice(
                data=mask, img=self.viewer.getImageItem()
            )
            mask[x_slice, y_slice] = True

        return mask


class ProcessingDialog(QtWidgets.QDialog):
    """
    Modal dialog used to select dataset processing options.
    """

    processing_parameters_signal = QtCore.pyqtSignal(dict)
    error_message_signal = QtCore.pyqtSignal(str)

    def __init__(self, raw, **kwargs):
        """
        Parameters
        ----------
        raw : AbstractRawDataset instance
        """
        super().__init__(**kwargs)
        self.setModal(True)
        self.setWindowTitle("Diffraction Dataset Processing")

        self.error_message_signal.connect(self.show_error_message)

        image = raw.raw_data(timedelay=raw.time_points[0], scan=raw.scans[0], bgr=True)
        self.mask_widget = MaskCreator(image, parent=self)

        title = QtWidgets.QLabel("<h2>Data Reduction Options<\\h2>")
        title.setTextFormat(QtCore.Qt.RichText)
        title.setAlignment(QtCore.Qt.AlignCenter)

        self.processes_widget = QtWidgets.QSpinBox(parent=self)
        self.processes_widget.setRange(1, cpu_count() - 1)
        self.processes_widget.setValue(min([cpu_count(), 7]))

        self.dtype_widget = QtWidgets.QComboBox(parent=self)
        self.dtype_widget.addItems(DTYPE_NAMES.keys())
        self.dtype_widget.setCurrentText("Auto")

        # Set exclude scan widget with a validator
        self.exclude_scans_widget = QtWidgets.QLineEdit(parent=self)
        self.exclude_scans_widget.setPlaceholderText("e.g. 1:5, 6, 7, 10:50, 100")

        self.alignment_tf_widget = QtWidgets.QCheckBox(
            "Perform alignment (?)", parent=self
        )
        self.alignment_tf_widget.setToolTip(alignment_help)
        self.alignment_tf_widget.setChecked(False)

        self.normalization_tf_widget = QtWidgets.QCheckBox("Normalize (?)", parent=self)
        self.normalization_tf_widget.setToolTip(normalization_help)
        self.normalization_tf_widget.setChecked(True)

        self.fletcher32_widget = QtWidgets.QCheckBox(
            "Enable Fletcher32 filter (?)", parent=self
        )
        self.fletcher32_widget.setToolTip(fletcher32_help)

        self.shuffle_filter_widget = QtWidgets.QCheckBox(
            "Enable shuffle filter (?)", parent=self
        )
        self.shuffle_filter_widget.setToolTip(shuffle_help)

        self.no_compression_btn = QtWidgets.QRadioButton("No compression", self)
        self.no_compression_btn.setChecked(True)

        self.lzf_btn = QtWidgets.QRadioButton("LZF", self)

        self.gzip_btn = QtWidgets.QRadioButton("GZIP", self)

        filter_btns = QtWidgets.QVBoxLayout()
        filter_btns.addWidget(self.no_compression_btn)
        filter_btns.addWidget(self.lzf_btn)
        filter_btns.addWidget(self.gzip_btn)

        self.filters = QtWidgets.QGroupBox("HDF5 Compression filters", parent=self)
        self.filters.setLayout(filter_btns)
        self.filters.setFlat(True)

        self.gzip_level_widget = QtWidgets.QSpinBox(self)
        self.gzip_level_widget.setRange(0, 9)
        self.gzip_level_widget.setValue(4)
        self.gzip_btn.toggled.connect(self.gzip_level_widget.setEnabled)
        self.gzip_level_widget.setEnabled(False)

        # mask controls
        self.add_rect_mask_btn = QtWidgets.QPushButton("Add rectangular mask", self)
        self.add_rect_mask_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        self.add_rect_mask_btn.clicked.connect(self.mask_widget.add_rect_mask)

        self.add_circ_mask_btn = QtWidgets.QPushButton("Add circular mask", self)
        self.add_circ_mask_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        self.add_circ_mask_btn.clicked.connect(self.mask_widget.add_circ_mask)

        self.preview_mask_btn = QtWidgets.QPushButton("Preview mask", self)
        self.preview_mask_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        self.preview_mask_btn.clicked.connect(self.mask_widget.show_preview_mask)

        self.clear_masks_btn = QtWidgets.QPushButton("Clear all masks", self)
        self.clear_masks_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        self.clear_masks_btn.clicked.connect(self.mask_widget.clear_masks)

        mask_btns = QtWidgets.QGridLayout()
        mask_btns.addWidget(self.add_circ_mask_btn, 0, 0)
        mask_btns.addWidget(self.add_rect_mask_btn, 0, 1)
        mask_btns.addWidget(self.preview_mask_btn, 1, 0)
        mask_btns.addWidget(self.clear_masks_btn, 1, 1)

        self.mask_controls = QtWidgets.QGroupBox("Mask controls", parent=self)
        self.mask_controls.setLayout(mask_btns)
        self.mask_controls.setFlat(True)

        save_btn = QtWidgets.QPushButton("Launch", self)
        save_btn.clicked.connect(self.accept)

        cancel_btn = QtWidgets.QPushButton("Cancel", self)
        cancel_btn.clicked.connect(self.reject)
        cancel_btn.setDefault(True)

        # Determine settings
        self.file_dialog = QtWidgets.QFileDialog(parent=self)

        processing_options = QtWidgets.QFormLayout()
        processing_options.addRow("Number of CPU cores:", self.processes_widget)
        processing_options.addRow("Scans to exclude: ", self.exclude_scans_widget)
        processing_options.addRow("Final data type: ", self.dtype_widget)
        processing_options.addRow(self.mask_controls)
        processing_options.addRow(self.alignment_tf_widget)
        processing_options.addRow(self.normalization_tf_widget)
        processing_options.addRow(self.fletcher32_widget)
        processing_options.addRow(self.shuffle_filter_widget)
        processing_options.addRow(self.filters)
        processing_options.addRow("GZIP level: ", self.gzip_level_widget)

        buttons = QtWidgets.QHBoxLayout()
        buttons.addWidget(save_btn)
        buttons.addWidget(cancel_btn)

        params_layout = QtWidgets.QVBoxLayout()
        params_layout.addWidget(title)
        params_layout.addSpacing(10)
        params_layout.addLayout(processing_options)
        params_layout.addLayout(buttons)

        params_widget = QtWidgets.QFrame()
        params_widget.setLayout(params_layout)
        params_widget.setFrameShadow(QtWidgets.QFrame.Sunken)
        params_widget.setFrameShape(QtWidgets.QFrame.Panel)
        params_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        right_layout = QtWidgets.QVBoxLayout()
        right_layout.addWidget(params_widget)
        right_layout.addStretch()

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self.mask_widget)
        layout.addLayout(right_layout)

        self.setLayout(layout)

    @QtCore.pyqtSlot(str)
    def show_error_message(self, msg):
        self.error_dialog = QtGui.QErrorMessage(parent=self)
        self.error_dialog.showMessage(msg)

    def file_params(self):
        """ Returns a dictionary with HDF5 file parameters """

        fletcher32 = self.fletcher32_widget.isChecked()
        shuffle = self.shuffle_filter_widget.isChecked()

        params = {"fletcher32": fletcher32, "shuffle": shuffle}

        if self.no_compression_btn.isChecked():
            return params

        params["compression"] = "lzf" if self.lzf_btn.isChecked() else "gzip"
        if params["compression"] == "gzip":
            params["compression_opts"] = self.gzip_level_widget.value()

        return params

    @QtCore.pyqtSlot()
    def accept(self):

        # If there is no mask, most probably the user has forgotten to add the mask
        if np.all(np.equal(self.mask_widget.composite_mask(), 0)):
            check_mask_box = QtWidgets.QMessageBox(parent=self)
            check_mask_box.setText(
                "Pixel masks are used through iris to enhance results. However, no mask was specified. \n\n Are you sure you want to continue?"
            )
            check_mask_box.setStandardButtons(
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No
            )
            check_mask_box.setDefaultButton(QtWidgets.QMessageBox.Cancel)
            res = check_mask_box.exec_()
            if res == QtWidgets.QMessageBox.No:
                return

        filename = self.file_dialog.getSaveFileName(filter="*.hdf5")[0]
        if filename == "":
            return

        # Force data type
        dtype = DTYPE_NAMES[self.dtype_widget.currentText()]

        exclude_scans_text = self.exclude_scans_widget.text()
        try:
            exclude_scans = parse_range(exclude_scans_text)
        except ValueError:
            self.error_message_signal.emit(
                f"Exclude scans unparseable:\n {exclude_scans_text}"
            )
            return

        # The arguments to the iris.processing.process function
        # more arguments will be added by controller
        kwargs = {
            "filename": filename,
            "processes": self.processes_widget.value(),
            "exclude_scans": exclude_scans,
            "dtype": dtype,
            "align": self.alignment_tf_widget.isChecked(),
            "normalize": self.normalization_tf_widget.isChecked(),
        }

        # Some parameters are from different widgets
        # HDF5 compression kwargs
        kwargs["valid_mask"] = np.logical_not(self.mask_widget.composite_mask())
        kwargs["ckwargs"] = self.file_params()

        self.processing_parameters_signal.emit(kwargs)
        super().accept()


def parse_range(range_str):
    """
    Parse an integer range into a list of numbers.

    Parameters
    ----------
    range_str : str
        String of the form : "-10, 1:5, 10:50, 100, 101".
        Ranges are inclusive (the endpoint is included). Can also be
        an empty string.

    Returns
    -------
    range : iterable of ints
        Iterable of integers (possibly empty). Guaranteed to be sorted and unique.

    Raises
    ------
    ValueError
        if the input ``range_str`` is unparseable.
    """
    range_str = str(range_str)
    range_str = range_str.replace(" ", "")
    if not range_str:
        return list()

    elements = range_str.split(",")
    if not elements:
        return list()

    iterable = list()

    # Two possibilities : ints or ranges
    # Either elem = int
    # or     elem = start:stop
    # Note : stop + 1 because inclusive ranges
    for elem in elements:
        try:
            fl = int(elem)
            iterable.append(fl)
        except ValueError:
            try:
                start, stop = tuple(map(int, elem.split(":")))
                iterable.extend(range(start, stop + 1))
            except:
                # Raise exception from None because full traceback is not useful
                # especially in terms of GUI error messages
                raise ValueError("Unparseable input: ", range_str) from None

    return list(sorted(set(iterable)))



# -*- coding: utf-8 -*-
"""
Time-series widget
"""

from functools import lru_cache

import numpy as np
from PyQt5 import QtCore, QtGui, QtWidgets
from pyqtgraph import PlotWidget, mkBrush, mkPen, TextItem

from skued import spectrum_colors, exponential

from scipy.optimize import curve_fit


@lru_cache(maxsize=1)
def pens_and_brushes(num):
    qcolors = tuple(map(lambda c: QtGui.QColor.fromRgbF(*c), spectrum_colors(num)))
    pens = list(map(mkPen, qcolors))
    brushes = list(map(mkBrush, qcolors))
    return pens, brushes


class TimeSeriesWidget(QtWidgets.QWidget):
    """ Time-series widget with built-in display controls. """

    # Internal refresh signal
    _refresh_signal = QtCore.pyqtSignal(np.ndarray, np.ndarray)

    # Signal to change the y-axis units
    _yaxis_units_signal = QtCore.pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.plot_widget = PlotWidget(
            parent=self,
            title="Diffraction time-series",
            labels={"left": ("Intensity", "a. u."), "bottom": ("time", "ps")},
        )
        self._yaxis_units_signal.connect(self.set_yaxis_units)

        self.plot_widget.getPlotItem().showGrid(x=True, y=True)

        # Internal state on whether or not to 'connect the dots'
        self._time_series_connect = False
        self._symbol_size = 5

        # Internal memory on last time-series info
        self._last_times = None
        self._last_intensities_abs = None
        self._refresh_signal.connect(self.plot)

        self.connect_widget = QtWidgets.QCheckBox("Connect time-series", self)
        self.connect_widget.toggled.connect(self.enable_connect)

        self.symbol_size_widget = QtWidgets.QSpinBox(parent=self)
        self.symbol_size_widget.setRange(1, 25)
        self.symbol_size_widget.setValue(self._symbol_size)
        self.symbol_size_widget.setPrefix("Symbol size: ")
        self.symbol_size_widget.valueChanged.connect(self.set_symbol_size)

        self.absolute_intensity_widget = QtWidgets.QCheckBox(
            "Show absolute intensity", self
        )
        self.absolute_intensity_widget.toggled.connect(self.toggle_absolute_intensity)

        self.exponential_fit_widget = QtWidgets.QPushButton(
            "Calculate exponential decay", self
        )
        self.exponential_fit_widget.clicked.connect(self.fit_exponential_decay)

        self.export_timeseries_widget = QtWidgets.QPushButton(
            "Export time-series data", self
        )
        self.export_timeseries_widget.clicked.connect(self.export_timeseries)

        self.fit_constants_label = QtWidgets.QLabel(self)

        self.controls = QtWidgets.QHBoxLayout()
        self.controls.addWidget(self.connect_widget)
        self.controls.addWidget(self.absolute_intensity_widget)
        self.controls.addWidget(self.symbol_size_widget)
        self.controls.addWidget(self.exponential_fit_widget)
        self.controls.addWidget(self.fit_constants_label)
        self.controls.addWidget(self.export_timeseries_widget)
        self.controls.addStretch(1)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.plot_widget)
        layout.addLayout(self.controls)
        self.setLayout(layout)

    @QtCore.pyqtSlot(str)
    def set_yaxis_units(self, units):
        """ Set the y-axis label """
        self.plot_widget.getPlotItem().setLabel("left", text="Intensity", units=units)

    @QtCore.pyqtSlot(bool)
    def toggle_absolute_intensity(self, absolute):
        # Normalize to intensity before time-zero
        if absolute:
            self._yaxis_units_signal.emit("counts")
        else:
            self._yaxis_units_signal.emit("a.u.")

        self.refresh()

    @QtCore.pyqtSlot(np.ndarray, np.ndarray)
    def plot(self, time_points, intensity):
        """
        Plot the a time-series. Time-series are normalized to the intensity before photoexcitation.

        Parameters
        ----------
        time_points : `~numpy.ndarray`, shape (N,)
            Time-delays [ps]
        intensity : `~numpy.ndarray`, shape (N,)
            Diffracted intensity in absolute units.
        """
        self._last_times = np.asarray(time_points)
        self._last_intensities_abs = np.asarray(intensity)

        # Normalize to intensity before time-zero
        # Note that the change in units is taken care of in the toggle_absolute_intensity method
        absolute = self.absolute_intensity_widget.isChecked()
        intensity = (
            self._last_intensities_abs
            if absolute
            else self._last_intensities_abs
            / np.mean(self._last_intensities_abs[self._last_times < 0])
        )

        # Only compute the colors if number of time-points changes or first time
        pens, brushes = pens_and_brushes(len(time_points))

        connect_kwargs = {"pen": None} if not self._time_series_connect else dict()
        self.plot_widget.plot(
            x=self._last_times,
            y=intensity,
            symbol="o",
            symbolPen=pens,
            symbolBrush=brushes,
            symbolSize=self._symbol_size,
            clear=True,
            **connect_kwargs,
        )

        # Don't forget to clear the fit constants
        self.fit_constants_label.clear()

    @QtCore.pyqtSlot()
    def fit_exponential_decay(self):
        """ Try to fit to a time-series with an exponential decay. If successful, plot the result. """
        times = self._last_times
        intensity = self._last_intensities_abs

        # time-zero, amplitude, time-constant, offset
        initial_guesses = (0, intensity.max(), 1, intensity.min())

        try:
            params, pcov = curve_fit(exponential, times, intensity, p0=initial_guesses)
        except RuntimeError:
            return

        fit = exponential(times, *params)
        absolute = self.absolute_intensity_widget.isChecked()
        if not absolute:
            fit /= np.mean(self._last_intensities_abs[self._last_times < 0])
        self.plot_widget.plot(x=times, y=fit, symbol=None, clear=False)

        # Write fit parameters to text items
        tconst, tconsterr = params[2], np.sqrt(pcov[2, 2])
        tzero, tzeroerr = params[0], np.sqrt(pcov[0, 0])
        self.fit_constants_label.setText(
            f"Time-constant: ({tconst:.3f}±{tconsterr:.3f} ps, Time-zero: ({tzero:.3f}±{tzeroerr:.3f}) ps"
        )

    @QtCore.pyqtSlot()
    def export_timeseries(self):
        """ Allow exporting the time-series data. """
        times = self._last_times
        intensity = self._last_intensities_abs

        if (times is None) or (intensity is None):
            return QtWidgets.QMessageBox.warning("No time-series to export.")

        file_dialog = QtWidgets.QFileDialog()
        path = file_dialog.getSaveFileName(
            parent=self, caption="Export time-series data", filter="*.csv"
        )[0]
        if not path:
            return

        data = np.empty(shape=(len(times), 2), dtype=float)
        data[:, 0] = times
        data[:, 1] = intensity
        np.savetxt(path, data, delimiter=",", header="times [ps], intensity [cnts]")

    @QtCore.pyqtSlot()
    def refresh(self):
        self._refresh_signal.emit(self._last_times, self._last_intensities_abs)

    @QtCore.pyqtSlot()
    def clear(self):
        self.plot_widget.clear()

    @QtCore.pyqtSlot(bool)
    def enable_connect(self, toggle):
        self._time_series_connect = toggle
        self.refresh()

    @QtCore.pyqtSlot(int)
    def set_symbol_size(self, size):
        self._symbol_size = size
        self.refresh()



# -*- coding: utf-8 -*-
__author__ = "Laurent P. René de Cotret"
__email__ = "laurent.renedecotret@mail.mcgill.ca"
__license__ = "GPLv3"
__version__ = "5.3.1"

from .raw import AbstractRawDataset, check_raw_bounds, open_raw
from .dataset import DiffractionDataset, MigrationWarning, MigrationError
from .powder import PowderDiffractionDataset
from .meta import ExperimentalParameter
from .plugins import install_plugin, load_plugin

from . import plugins



# -*- coding: utf-8 -*-
"""
Control bar for all Iris's controls
"""
from collections import Iterable

from PyQt5 import QtCore, QtGui, QtWidgets
from pywt import Modes

from skued import available_dt_filters, available_first_stage_filters


class ControlBar(QtWidgets.QWidget):

    raw_data_request = QtCore.pyqtSignal(int, int)  # timedelay index, scan
    averaged_data_request = QtCore.pyqtSignal(int)  # timedelay index

    baseline_computation_parameters = QtCore.pyqtSignal(dict)
    time_zero_shift = QtCore.pyqtSignal(float)
    notes_updated = QtCore.pyqtSignal(str)
    enable_connect_time_series = QtCore.pyqtSignal(bool)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.raw_dataset_controls = RawDatasetControl(parent=self)
        self.raw_dataset_controls.timedelay_widget.valueChanged.connect(
            self.request_raw_data
        )
        self.raw_dataset_controls.scan_widget.valueChanged.connect(
            self.request_raw_data
        )
        #        self.raw_dataset_controls.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)

        self.diffraction_dataset_controls = DiffractionDatasetControl(parent=self)
        self.diffraction_dataset_controls.timedelay_widget.valueChanged.connect(
            self.averaged_data_request
        )
        self.diffraction_dataset_controls.time_zero_shift_widget.editingFinished.connect(
            self.shift_time_zero
        )
        self.diffraction_dataset_controls.clear_time_zero_shift_btn.clicked.connect(
            lambda _: self.time_zero_shift.emit(0)
        )
        #        self.diffraction_dataset_controls.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)

        self.powder_diffraction_dataset_controls = PowderDiffractionDatasetControl(
            parent=self
        )
        self.powder_diffraction_dataset_controls.compute_baseline_btn.clicked.connect(
            self.request_baseline_computation
        )
        #        self.powder_diffraction_dataset_controls.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)

        self.metadata_widget = MetadataWidget(parent=self)

        self.notes_editor = NotesEditor(parent=self)
        self.notes_editor.notes_updated.connect(self.notes_updated)

        self.metadata_and_notes_widget = QtWidgets.QFrame(parent=self)
        _layout = QtWidgets.QVBoxLayout()
        _layout.addWidget(self.metadata_widget)
        _layout.addWidget(self.notes_editor)
        self.metadata_and_notes_widget.setLayout(_layout)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.raw_dataset_controls)
        layout.addWidget(self.diffraction_dataset_controls)
        layout.addWidget(self.powder_diffraction_dataset_controls)
        layout.addWidget(self.metadata_and_notes_widget)
        layout.addStretch()
        self.setLayout(layout)

        for frame in (
            self.raw_dataset_controls,
            self.diffraction_dataset_controls,
            self.powder_diffraction_dataset_controls,
            self.metadata_and_notes_widget,
        ):
            frame.setFrameShadow(QtWidgets.QFrame.Sunken)
            frame.setFrameShape(QtWidgets.QFrame.Panel)

        self.setMaximumWidth(self.notes_editor.maximumWidth())

    @QtCore.pyqtSlot(dict)
    def update_raw_dataset_metadata(self, metadata):
        self.raw_dataset_controls.update_dataset_metadata(metadata)

    @QtCore.pyqtSlot(dict)
    def update_dataset_metadata(self, metadata):
        self.diffraction_dataset_controls.update_dataset_metadata(metadata)
        self.notes_editor.editor.setPlainText(
            metadata.pop("notes", "No notes available")
        )
        self.metadata_widget.set_metadata(metadata)

    @QtCore.pyqtSlot(int)
    def request_raw_data(self, wtv):
        self.raw_data_request.emit(
            self.raw_dataset_controls.timedelay_widget.value(),
            self.raw_dataset_controls.scan_widget.value() + 1,
        )  # scans are numbered starting at 1

    @QtCore.pyqtSlot()
    def request_baseline_computation(self):
        self.baseline_computation_parameters.emit(
            self.powder_diffraction_dataset_controls.baseline_parameters()
        )

    @QtCore.pyqtSlot()
    def shift_time_zero(self):
        shift = self.diffraction_dataset_controls.time_zero_shift_widget.value()
        self.time_zero_shift.emit(shift)


class RawDatasetControl(QtWidgets.QFrame):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        #############################
        # Navigating through raw data
        self.td_label = QtWidgets.QLabel("Time-delay: ")
        self.td_label.setAlignment(QtCore.Qt.AlignCenter)
        self.timedelay_widget = QtWidgets.QSlider(QtCore.Qt.Horizontal, parent=self)
        self.timedelay_widget.setMinimum(0)
        self.timedelay_widget.setTracking(False)
        self.timedelay_widget.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.timedelay_widget.setTickInterval(1)
        self.timedelay_widget.sliderMoved.connect(
            lambda pos: self.td_label.setText(
                f"Time-delay: {self.time_points[pos]:.3f}ps"
            )
        )

        self.s_label = QtWidgets.QLabel("Scan: ")
        self.s_label.setAlignment(QtCore.Qt.AlignCenter)
        self.scan_widget = QtWidgets.QSlider(QtCore.Qt.Horizontal, parent=self)
        self.scan_widget.setMinimum(0)
        self.scan_widget.setTracking(False)
        self.scan_widget.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.scan_widget.setTickInterval(1)
        self.scan_widget.sliderMoved.connect(
            lambda pos: self.s_label.setText(f"Scan: {self.scans[pos]:d}")
        )

        prev_timedelay_btn = QtWidgets.QPushButton("<", self)
        prev_timedelay_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        prev_timedelay_btn.clicked.connect(self.goto_prev_timedelay)

        next_timedelay_btn = QtWidgets.QPushButton(">", self)
        next_timedelay_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        next_timedelay_btn.clicked.connect(self.goto_next_timedelay)

        prev_scan_btn = QtWidgets.QPushButton("<", self)
        prev_scan_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        prev_scan_btn.clicked.connect(self.goto_prev_scan)

        next_scan_btn = QtWidgets.QPushButton(">", self)
        next_scan_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        next_scan_btn.clicked.connect(self.goto_next_scan)

        time_layout = QtWidgets.QHBoxLayout()
        time_layout.addWidget(self.td_label)
        time_layout.addWidget(self.timedelay_widget)
        time_layout.addWidget(prev_timedelay_btn)
        time_layout.addWidget(next_timedelay_btn)

        scan_layout = QtWidgets.QHBoxLayout()
        scan_layout.addWidget(self.s_label)
        scan_layout.addWidget(self.scan_widget)
        scan_layout.addWidget(prev_scan_btn)
        scan_layout.addWidget(next_scan_btn)

        sliders = QtWidgets.QVBoxLayout()
        sliders.addLayout(time_layout)
        sliders.addLayout(scan_layout)

        title = QtWidgets.QLabel("<h2>Raw dataset controls<\\h2>")
        title.setTextFormat(QtCore.Qt.RichText)
        title.setAlignment(QtCore.Qt.AlignCenter)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(title)
        layout.addLayout(sliders)
        self.setLayout(layout)
        self.resize(self.minimumSize())

    def update_dataset_metadata(self, metadata):
        self.time_points = metadata.get("time_points")
        self.scans = metadata.get("scans")

        self.timedelay_widget.setRange(0, len(self.time_points) - 1)
        self.scan_widget.setRange(0, len(self.scans) - 1)
        self.timedelay_widget.triggerAction(5)
        self.timedelay_widget.sliderMoved.emit(0)
        self.scan_widget.triggerAction(5)
        self.scan_widget.sliderMoved.emit(0)

    @QtCore.pyqtSlot()
    def goto_prev_timedelay(self):
        self.timedelay_widget.setSliderDown(True)
        self.timedelay_widget.triggerAction(2)
        self.timedelay_widget.setSliderDown(False)

    @QtCore.pyqtSlot()
    def goto_next_timedelay(self):
        self.timedelay_widget.setSliderDown(True)
        self.timedelay_widget.triggerAction(1)
        self.timedelay_widget.setSliderDown(False)

    @QtCore.pyqtSlot()
    def goto_prev_scan(self):
        self.scan_widget.setSliderDown(True)
        self.scan_widget.triggerAction(2)
        self.scan_widget.setSliderDown(False)

    @QtCore.pyqtSlot()
    def goto_next_scan(self):
        self.scan_widget.setSliderDown(True)
        self.scan_widget.triggerAction(1)
        self.scan_widget.setSliderDown(False)


class DiffractionDatasetControl(QtWidgets.QFrame):

    time_zero_shift = QtCore.pyqtSignal(float)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        ################################
        # Diffraction dataset navigation
        self.td_label = QtWidgets.QLabel("Time-delay: ")
        self.td_label.setAlignment(QtCore.Qt.AlignCenter)
        self.timedelay_widget = QtWidgets.QSlider(QtCore.Qt.Horizontal, parent=self)
        self.timedelay_widget.setMinimum(0)
        self.timedelay_widget.setTracking(False)
        self.timedelay_widget.setTickPosition(QtWidgets.QSlider.TicksBelow)
        self.timedelay_widget.setTickInterval(1)
        self.timedelay_widget.sliderMoved.connect(
            lambda pos: self.td_label.setText(
                f"Time-delay: {self.time_points[pos]:.3f}ps"
            )
        )

        # Time-zero shift control
        # QDoubleSpinbox does not have a slot that sets the value without notifying everybody
        self.time_zero_shift_widget = QtWidgets.QDoubleSpinBox(parent=self)
        self.time_zero_shift_widget.setRange(-1000, 1000)
        self.time_zero_shift_widget.setDecimals(3)
        self.time_zero_shift_widget.setSingleStep(0.5)
        self.time_zero_shift_widget.setSuffix(" ps")
        self.time_zero_shift_widget.setValue(0.0)

        self.clear_time_zero_shift_btn = QtWidgets.QPushButton(
            "Clear time-zero shift", parent=self
        )
        self.clear_time_zero_shift_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        prev_btn = QtWidgets.QPushButton("<", self)
        prev_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        prev_btn.clicked.connect(self.goto_prev)

        next_btn = QtWidgets.QPushButton(">", self)
        next_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        next_btn.clicked.connect(self.goto_next)

        sliders = QtWidgets.QHBoxLayout()
        sliders.addWidget(self.td_label)
        sliders.addWidget(self.timedelay_widget)
        sliders.addWidget(prev_btn)
        sliders.addWidget(next_btn)

        time_zero_shift_layout = QtWidgets.QHBoxLayout()
        time_zero_shift_layout.addWidget(
            QtWidgets.QLabel("Time-zero shift: ", parent=self)
        )
        time_zero_shift_layout.addWidget(self.time_zero_shift_widget)
        time_zero_shift_layout.addWidget(self.clear_time_zero_shift_btn)

        title = QtWidgets.QLabel("<h2>Diffraction dataset controls<\\h2>")
        title.setTextFormat(QtCore.Qt.RichText)
        title.setAlignment(QtCore.Qt.AlignCenter)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(title)
        layout.addLayout(sliders)
        layout.addLayout(time_zero_shift_layout)
        self.setLayout(layout)
        self.resize(self.minimumSize())

    def update_dataset_metadata(self, metadata):
        self.time_points = metadata.get("time_points")
        t0_shift = metadata.get("time_zero_shift")

        self.timedelay_widget.setRange(0, len(self.time_points) - 1)
        self.timedelay_widget.triggerAction(5)  # SliderToMinimum
        self.timedelay_widget.sliderMoved.emit(0)

        if t0_shift != self.time_zero_shift_widget.value():
            self.time_zero_shift_widget.setValue(t0_shift)

    @QtCore.pyqtSlot()
    def goto_prev(self):
        self.timedelay_widget.setSliderDown(True)
        self.timedelay_widget.triggerAction(2)
        self.timedelay_widget.setSliderDown(False)

    @QtCore.pyqtSlot()
    def goto_next(self):
        self.timedelay_widget.setSliderDown(True)
        self.timedelay_widget.triggerAction(1)
        self.timedelay_widget.setSliderDown(False)


class PowderDiffractionDatasetControl(QtWidgets.QFrame):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        ######################
        # baseline computation
        self.first_stage_cb = QtWidgets.QComboBox()
        self.first_stage_cb.addItems(available_first_stage_filters())
        if "sym6" in available_first_stage_filters():
            self.first_stage_cb.setCurrentText("sym6")

        self.wavelet_cb = QtWidgets.QComboBox()
        self.wavelet_cb.addItems(available_dt_filters())
        if "qshift3" in available_dt_filters():
            self.wavelet_cb.setCurrentText("qshift3")

        self.mode_cb = QtWidgets.QComboBox()
        self.mode_cb.addItems(Modes.modes)
        if "smooth" in Modes.modes:
            self.mode_cb.setCurrentText("constant")

        self.max_iter_widget = QtWidgets.QSpinBox()
        self.max_iter_widget.setRange(0, 1000)
        self.max_iter_widget.setValue(100)

        self.compute_baseline_btn = QtWidgets.QPushButton(
            "Compute baseline", parent=self
        )
        self.compute_baseline_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        baseline_controls = QtWidgets.QFormLayout()
        baseline_controls.addRow("First stage wavelet: ", self.first_stage_cb)
        baseline_controls.addRow("Dual-tree wavelet: ", self.wavelet_cb)
        baseline_controls.addRow("Extensions mode: ", self.mode_cb)
        baseline_controls.addRow("Iterations: ", self.max_iter_widget)
        baseline_controls.addWidget(self.compute_baseline_btn)

        baseline_computation = QtWidgets.QGroupBox(
            title="Baseline parameters", parent=self
        )
        baseline_computation.setLayout(baseline_controls)

        # TODO: add callback and progressbar for computing the baseline?

        title = QtWidgets.QLabel("<h2>Powder dataset controls<\\h2>")
        title.setTextFormat(QtCore.Qt.RichText)
        title.setAlignment(QtCore.Qt.AlignCenter)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(title)
        layout.addWidget(baseline_computation)
        self.setLayout(layout)
        self.resize(self.minimumSize())

    def baseline_parameters(self):
        """ Returns a dictionary of baseline-computation parameters """
        return {
            "first_stage": self.first_stage_cb.currentText(),
            "wavelet": self.wavelet_cb.currentText(),
            "mode": self.mode_cb.currentText(),
            "max_iter": self.max_iter_widget.value(),
            "level": None,
        }


class MetadataWidget(QtWidgets.QWidget):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        title = QtWidgets.QLabel("<h2>Dataset metadata<\\h2>", parent=self)
        title.setTextFormat(QtCore.Qt.RichText)
        title.setAlignment(QtCore.Qt.AlignCenter)

        self.table = QtWidgets.QTableWidget(parent=self)
        self.table.setColumnCount(2)
        self.table.horizontalHeader().hide()
        self.table.verticalHeader().hide()
        self.table.setEditTriggers(
            QtWidgets.QAbstractItemView.NoEditTriggers
        )  # no edit triggers, see QAbstractItemViews

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(title)
        layout.addWidget(self.table)
        self.setLayout(layout)
        self.resize(self.minimumSize())

    @QtCore.pyqtSlot(dict)
    def set_metadata(self, metadata):
        self.table.clear()
        self.table.setRowCount(
            len(metadata) - 1 if "notes" in metadata else len(metadata)
        )
        for row, (key, value) in enumerate(sorted(metadata.items())):
            if isinstance(value, Iterable) and (not isinstance(value, str)):
                if len(value) > 4:
                    key += " (length)"
                    value = len(tuple(value))

            value = str(value)
            # We show tool tip on hover because
            # some values are long (e.g. filenames)
            key_item = QtWidgets.QTableWidgetItem(key)
            key_item.setToolTip(key)

            value_item = QtWidgets.QTableWidgetItem(value)
            value_item.setToolTip(value)

            self.table.setItem(row, 0, key_item)
            self.table.setItem(row, 1, value_item)

        self.table.resizeColumnsToContents()
        self.table.horizontalHeader().setStretchLastSection(True)


class NotesEditor(QtWidgets.QFrame):

    notes_updated = QtCore.pyqtSignal(str)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        title = QtWidgets.QLabel("<h2>Dataset notes and remarks<\\h2>", parent=self)
        title.setTextFormat(QtCore.Qt.RichText)
        title.setAlignment(QtCore.Qt.AlignCenter)

        update_btn = QtWidgets.QPushButton("Update notes", self)
        update_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        update_btn.clicked.connect(self.update_notes)

        self.editor = QtWidgets.QTextEdit(parent=self)

        # Set editor size such that 60 characters will fit
        font_info = QtGui.QFontInfo(self.editor.currentFont())
        self.editor.setMaximumWidth(40 * font_info.pixelSize())
        self.editor.setLineWrapMode(QtWidgets.QTextEdit.WidgetWidth)  # widget width

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(title)
        layout.addWidget(self.editor)
        layout.addWidget(update_btn, 1, QtCore.Qt.AlignHCenter)
        self.setLayout(layout)
        self.setMaximumWidth(self.editor.maximumWidth())
        self.resize(self.minimumSize())

    @QtCore.pyqtSlot()
    def update_notes(self):
        self.notes_updated.emit(self.editor.toPlainText())



# -*- coding: utf-8 -*-
"""
Dialog for editing a restricted set of metadata
"""
from PyQt5 import QtCore, QtWidgets


class MetadataEditDialog(QtWidgets.QDialog):
    """Modal dialog to specify modify metadata
    like camera distance, etc."""

    updated_metadata_signal = QtCore.pyqtSignal(dict)

    def __init__(self, config, *args, **kwargs):
        """
        Parameters
        ----------
        config : dict
            Dictionary of the currently-saved experimental setup.
        """
        super().__init__(*args, **kwargs)
        self.setModal(True)
        self.setWindowTitle("Experimental set-up")

        self.camera_distance_widget = QtWidgets.QDoubleSpinBox(parent=self)
        self.camera_distance_widget.setRange(0, 100)
        self.camera_distance_widget.setDecimals(4)
        self.camera_distance_widget.setSingleStep(0.0001)
        self.camera_distance_widget.setSuffix(" m")
        self.camera_distance_widget.setValue(float(config["camera_length"]))

        self.pixel_width_widget = QtWidgets.QDoubleSpinBox(parent=self)
        self.pixel_width_widget.setRange(0, 1000)
        self.pixel_width_widget.setDecimals(2)
        self.pixel_width_widget.setSingleStep(0.01)
        self.pixel_width_widget.setSuffix(" μm")
        self.pixel_width_widget.setValue(float(config["pixel_width"]) * 1e6)

        self.temperature_widget = QtWidgets.QDoubleSpinBox(parent=self)
        self.temperature_widget.setRange(0, 1e4)
        self.temperature_widget.setDecimals(2)
        self.temperature_widget.setSingleStep(1)
        self.temperature_widget.setSuffix(" K")
        self.temperature_widget.setValue(float(config["temperature"]))

        self.fluence_widget = QtWidgets.QDoubleSpinBox(parent=self)
        self.fluence_widget.setRange(0, 1000)
        self.fluence_widget.setDecimals(2)
        self.fluence_widget.setSingleStep(1)
        self.fluence_widget.setSuffix(" mJ/cm²")
        self.fluence_widget.setValue(float(config["fluence"]))

        self.accept_btn = QtWidgets.QPushButton("Confirm", self)
        self.accept_btn.clicked.connect(self.accept)

        self.cancel_btn = QtWidgets.QPushButton("Cancel", self)
        self.cancel_btn.clicked.connect(self.reject)
        self.cancel_btn.setDefault(True)

        btns = QtWidgets.QHBoxLayout()
        btns.addWidget(self.accept_btn)
        btns.addWidget(self.cancel_btn)

        params = QtWidgets.QFormLayout()
        params.addRow("Camera distance: ", self.camera_distance_widget)
        params.addRow("Pixel width: ", self.pixel_width_widget)
        params.addRow("Temperature: ", self.temperature_widget)
        params.addRow("Fluence: ", self.fluence_widget)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addLayout(params)
        self.layout.addLayout(btns)
        self.setLayout(self.layout)

    @QtCore.pyqtSlot()
    def accept(self):
        params = {
            "camera_length": float(self.camera_distance_widget.value()),
            "pixel_width": float(self.pixel_width_widget.value()) * 1e-6,
            "temperature": float(self.temperature_widget.value()),
            "fluence": float(self.fluence_widget.value()),
        }

        self.updated_metadata_signal.emit(params)
        super().accept()



# -*- coding: utf-8 -*-
"""
Dialog for Q-vector calibration of powder diffraction data
"""
import numpy as np
import pyqtgraph as pg
from PyQt5 import QtCore, QtGui, QtWidgets

from crystals import Crystal
from skued import powder_calq


EXPLANATION = """Calibrate the scattering vector range of polycrystalline data. Select two peaks of known Miller indices by dragging the vertical lines, 
and use an appropriate structure. Some structures are built-in, but you can also use a CIF of your own. Make sure the structure 
parameters are what you expect before calibration. """.replace(
    "\n", ""
)


class MillerIndexWidget(QtWidgets.QWidget):
    """
    Widget for specifying a peak's Miller indices
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.h_widget = QtWidgets.QSpinBox(parent=self)
        self.h_widget.setPrefix("h: ")
        self.h_widget.setRange(-999, 999)
        self.h_widget.setValue(0)
        self.h_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        self.k_widget = QtWidgets.QSpinBox(parent=self)
        self.k_widget.setPrefix("k: ")
        self.k_widget.setRange(-999, 999)
        self.k_widget.setValue(0)
        self.k_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        self.l_widget = QtWidgets.QSpinBox(parent=self)
        self.l_widget.setPrefix("l: ")
        self.l_widget.setRange(-999, 999)
        self.l_widget.setValue(0)
        self.l_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.h_widget)
        self.layout.addWidget(self.k_widget)
        self.layout.addWidget(self.l_widget)

        self.setLayout(self.layout)

    @property
    def miller_indices(self):
        return self.h_widget.value(), self.k_widget.value(), self.l_widget.value()


class QCalibratorDialog(QtWidgets.QDialog):
    """
    Calibrate the scattering vector range from a polycrystalline diffraction pattern.

    Parameters
    ----------
    q : `~numpy.ndarray`
        Scattering vector array.
    I : `~numpy.ndarray`
        Powder diffraction pattern defined on the vector ``q``.
    """

    error_message = QtCore.pyqtSignal(str)
    new_crystal = QtCore.pyqtSignal(Crystal)
    calibration_parameters = QtCore.pyqtSignal(dict)

    def __init__(self, I, **kwargs):
        super().__init__(**kwargs)
        self.setModal(True)
        self.setWindowTitle("Scattering vector range calibration")

        title = QtWidgets.QLabel("<h2>Scattering Calibration<\\h2>")
        title.setTextFormat(QtCore.Qt.RichText)
        title.setAlignment(QtCore.Qt.AlignCenter)

        explanation_label = QtWidgets.QLabel(EXPLANATION)
        explanation_label.setAlignment(QtCore.Qt.AlignCenter)
        explanation_label.setWordWrap(True)

        self.intensity = I
        self.crystal = None

        self.error_message.connect(self.show_error_message)

        plot_widget = pg.PlotWidget(parent=self)
        plot_widget.plot(np.arange(0, len(self.intensity)), self.intensity)

        self.peak1_indicator = pg.InfiniteLine(0, movable=True)
        self.peak2_indicator = pg.InfiniteLine(len(I), movable=True)

        plot_widget.addItem(self.peak1_indicator)
        plot_widget.addItem(self.peak2_indicator)

        # Crystal creation ----------------------------------------------------
        database_widget = QtWidgets.QComboBox(parent=self)
        database_widget.addItems(sorted(Crystal.builtins))
        database_widget.currentTextChanged.connect(self.create_database_crystal)

        structure_file_btn = QtWidgets.QPushButton("Open explorer", self)
        structure_file_btn.clicked.connect(self.load_cif)
        structure_file_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        crystal_label = QtWidgets.QLabel(parent=self)
        crystal_label.setText("No structure selected.")
        crystal_label.setWordWrap(True)
        self.new_crystal.connect(
            lambda c: crystal_label.setText(str(c))
        )  # Use str, not repr, for shorter representation

        crystal_label_title = QtWidgets.QLabel("<h3>Selected crystal structure</h3>")
        crystal_label_title.setAlignment(QtCore.Qt.AlignHCenter)

        crystal_description_layout = QtWidgets.QVBoxLayout()
        crystal_description_layout.addWidget(crystal_label_title)
        crystal_description_layout.addWidget(crystal_label)

        crystal_description_widget = QtWidgets.QFrame(self)
        crystal_description_widget.setLayout(crystal_description_layout)
        crystal_description_widget.setFrameShadow(QtWidgets.QFrame.Sunken)
        crystal_description_widget.setFrameShape(QtWidgets.QFrame.Panel)

        crystal_creation_layout = QtWidgets.QFormLayout()
        crystal_creation_layout.addRow("Select a database structure: ", database_widget)
        crystal_creation_layout.addRow("Load structure from file: ", structure_file_btn)
        crystal_creation_layout.addRow(crystal_description_widget)

        # Peak specifications -------------------------------------------------
        self.left_peak_miller = MillerIndexWidget(parent=self)
        self.right_peak_miller = MillerIndexWidget(parent=self)

        peak_layout = QtWidgets.QFormLayout()
        peak_layout.addRow("Left peak: ", self.left_peak_miller)
        peak_layout.addRow("Right peak: ", self.right_peak_miller)
        peak_layout.setFormAlignment(QtCore.Qt.AlignCenter)

        self.accept_btn = QtWidgets.QPushButton("Calibrate", parent=self)
        self.accept_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        self.accept_btn.clicked.connect(self.accept)

        self.cancel_btn = QtWidgets.QPushButton("Cancel", parent=self)
        self.cancel_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        self.cancel_btn.clicked.connect(self.reject)
        self.cancel_btn.setDefault(True)

        btns = QtWidgets.QHBoxLayout()
        btns.addWidget(self.accept_btn)
        btns.addWidget(self.cancel_btn)

        params_layout = QtWidgets.QVBoxLayout()
        params_layout.addWidget(title)
        params_layout.addWidget(explanation_label)
        params_layout.addLayout(crystal_creation_layout)
        params_layout.addLayout(peak_layout)
        params_layout.addStretch()
        params_layout.addLayout(btns)

        params_widget = QtWidgets.QFrame(parent=self)
        params_widget.setLayout(params_layout)
        params_widget.setFrameShadow(QtWidgets.QFrame.Sunken)
        params_widget.setFrameShape(QtWidgets.QFrame.Panel)
        params_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        right_layout = QtWidgets.QVBoxLayout()
        right_layout.addWidget(params_widget)
        right_layout.addStretch()

        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(plot_widget)
        self.layout.addLayout(right_layout)
        self.setLayout(self.layout)

    @QtCore.pyqtSlot(str)
    def create_database_crystal(self, name):
        crystal = Crystal.from_database(name)
        self.crystal = crystal
        self.new_crystal.emit(self.crystal)

    @QtCore.pyqtSlot()
    def load_cif(self):
        path, *_ = QtWidgets.QFileDialog.getOpenFileName(
            parent=self, caption="Load structure from CIF", filter="*.cif"
        )
        if not path:
            return

        self.crystal = Crystal.from_cif(path)
        self.new_crystal.emit(self.crystal)

    @QtCore.pyqtSlot(str)
    def show_error_message(self, msg):
        self.error_dialog = QtGui.QErrorMessage(parent=self)
        self.error_dialog.showMessage(msg)

    @QtCore.pyqtSlot()
    def accept(self):
        if self.crystal is None:
            self.show_error_message(
                "Missing structure. Select a Crystal from the database or load a structure file (CIF)."
            )
            return

        positions = self.peak1_indicator.getXPos(), self.peak2_indicator.getXPos()
        left, right = min(positions), max(positions)

        params = {
            "crystal": self.crystal,
            "peak_indices": (int(left), int(right)),
            "miller_indices": (
                self.left_peak_miller.miller_indices,
                self.right_peak_miller.miller_indices,
            ),
        }
        self.calibration_parameters.emit(params)
        super().accept()



# -*- coding: utf-8 -*-
"""
Viewer widgets for DiffractionDatasets
"""
import pyqtgraph as pg
from PyQt5 import QtCore, QtWidgets
import numpy as np

from .time_series_widget import TimeSeriesWidget


class ProcessedDataViewer(QtWidgets.QWidget):
    """
    Widget displaying DiffractionDatasets
    """

    timeseries_rect_signal = QtCore.pyqtSignal(tuple)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.image_viewer = pg.ImageView(parent=self)
        self.image_viewer.setPredefinedGradient(
            "inferno"
        )  # to distinguish between raw and processed

        self.time_series_widget = TimeSeriesWidget(parent=self)

        self.cursor_info_widget = QtWidgets.QLabel(parent=self)
        self.cursor_info_widget.setAlignment(QtCore.Qt.AlignHCenter)

        self.timeseries_rect_region = pg.ROI(
            pos=[0, 0], size=[200, 200], pen=pg.mkPen("r")
        )
        self.timeseries_rect_region.addScaleHandle([1, 1], [0, 0])
        self.timeseries_rect_region.addScaleHandle([0, 0], [1, 1])

        self.roi_topleft_text = pg.TextItem("", anchor=(1, 1))
        self.roi_bottomright_text = pg.TextItem("", anchor=(0, 0))
        self.image_viewer.addItem(self.roi_topleft_text)
        self.image_viewer.addItem(self.roi_bottomright_text)

        # Signal proxies allow to rate-limit them
        self.__cursor_proxy = pg.SignalProxy(
            self.image_viewer.scene.sigMouseMoved,
            rateLimit=60,
            slot=self.update_cursor_info,
        )
        self.__dynamics_roi_proxy = pg.SignalProxy(
            self.timeseries_rect_region.sigRegionChanged,
            rateLimit=30,
            slot=self.update_timeseries_rect,
        )

        self.__bragg_peak_items = list()

        self.image_viewer.getView().addItem(self.timeseries_rect_region)
        self.timeseries_rect_region.hide()
        self.time_series_widget.hide()

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.image_viewer)
        self.layout.addWidget(self.cursor_info_widget)
        self.layout.addWidget(self.time_series_widget)
        self.setLayout(self.layout)

    @QtCore.pyqtSlot()
    def update_timeseries_rect(self):
        rect = self.timeseries_rect_region.parentBounds().toRect()

        # Add text to show rect bounds
        # If coordinate is negative, return 0
        # Note that in case pg.getConfigOption('imageAxisOrder') is not row-major,
        # The following bounds will not work
        x1 = round(max(0, rect.topLeft().y()))
        x2 = round(max(0, rect.y() + rect.height()))
        y1 = round(max(0, rect.topLeft().x()))
        y2 = round(max(0, rect.x() + rect.width()))

        self.timeseries_rect_signal.emit((x1, x2, y1, y2))

    @QtCore.pyqtSlot(object)
    def update_cursor_info(self, event):
        """ Provide information about the cursor (position, value, etc) """
        mouse_point = self.image_viewer.getView().mapSceneToView(event[0])
        i, j = int(mouse_point.x()), int(mouse_point.y())
        try:
            val = self.image_viewer.getImageItem().image[j, i]
        except IndexError:
            val = 0
        self.cursor_info_widget.setText(
            f"Position: ({i},{j}) | Pixel value: {val:.2f} cnts"
        )

    @QtCore.pyqtSlot(bool)
    def toggle_peak_dynamics(self, toggle):
        """ Toggle interactive peak dynamics region-of-interest"""
        if toggle:
            self.timeseries_rect_region.show()
        else:
            self.timeseries_rect_region.hide()
            self.roi_topleft_text.setText("")
            self.roi_bottomright_text.setText("")

        self.time_series_widget.setVisible(toggle)

    @QtCore.pyqtSlot(bool)
    def toggle_roi_bounds_text(self, enable):
        """ Toggle showing array indices around the peak dynamics region-of-interest """
        if enable:
            self.timeseries_rect_signal.connect(self._update_roi_bounds_text)
            self.update_timeseries_rect()
        else:
            self.roi_topleft_text.setText("")
            self.roi_bottomright_text.setText("")
            self.timeseries_rect_signal.disconnect(self._update_roi_bounds_text)

    @QtCore.pyqtSlot(tuple)
    def _update_roi_bounds_text(self, rect):
        """ Update the ROI bounds text based on the bounds in ``rect`` """
        x1, x2, y1, y2 = rect
        self.roi_topleft_text.setPos(y1, x1)
        self.roi_topleft_text.setText(f"({y1},{x1})")

        self.roi_bottomright_text.setPos(y2, x2)
        self.roi_bottomright_text.setText(f"({y2},{x2})")

    @QtCore.pyqtSlot(object, bool)
    def display(self, image, autocontrast=False):
        """
        Display an image in the form of an ndarray.

        Parameters
        ----------
        image : ndarray or None
            If None, the display is cleared.
        autocontrast: bool, optional
            If True, the image contrast will be adjusted
        """
        if image is None:
            self.image_viewer.clear()
            return

        self.image_viewer.setImage(
            image, autoLevels=autocontrast, autoRange=autocontrast
        )
        self.update_timeseries_rect()

    @QtCore.pyqtSlot(object, object)
    def display_peak_dynamics(self, times, intensities):
        """
        Display the time series associated with the integral between the bounds
        of the ROI
        """
        self.time_series_widget.plot(times, intensities)

    @QtCore.pyqtSlot(list)
    def display_bragg_peaks(self, peaks):
        """
        Highlight the location of Bragg peaks.

        Parameters
        ----------
        peaks : list of 2-tuple
            List of tuples [row, col] where Bragg peaks are located.
        """
        for item in self.__bragg_peak_items:
            self.image_viewer.removeItem(item)
        self.__bragg_peak_items.clear()

        for r, c in peaks:
            self.__bragg_peak_items.append(
                pg.CircleROI(
                    pos=(c, r), size=50, movable=False, resizable=False, removable=False
                )
            )

        for item in self.__bragg_peak_items:
            self.image_viewer.addItem(item)



# -*- coding: utf-8 -*-
"""
    Modified from:
        Author: Jared P. Sutton <jpsutton@gmail.com>
        License: LGPL
        Note: I've licensed this code as LGPL because it was a complete translation of the code found here...
        https://github.com/mojocorp/QProgressIndicator
"""

from PyQt5 import QtWidgets, QtCore, QtGui


class QBusyIndicator(QtWidgets.QWidget):
    def __init__(self, **kwargs):

        super().__init__(**kwargs)

        self.delay = property(int, self.animationDelay, self.setAnimationDelay)
        self.color = property(QtGui.QColor, self.getColor, self.setColor)

        self.m_angle = 0
        self.m_timerId = -1
        self.m_delay = 40
        self.m_displayedWhenStopped = True
        self.m_color = QtCore.Qt.white

        self.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self.setFocusPolicy(QtCore.Qt.NoFocus)
        self.show()

    @QtCore.pyqtSlot(bool)
    def toggle_animation(self, busy):
        if busy:
            self.m_angle = 0

            if self.m_timerId == -1:
                self.m_timerId = self.startTimer(self.m_delay)
        else:
            if self.m_timerId != -1:
                self.killTimer(self.m_timerId)

            self.m_timerId = -1
            self.update()

    def sizeHint(self):
        return QtCore.QSize(20, 20)

    def animationDelay(self):
        return self.delay

    def isAnimated(self):
        return self.m_timerId != -1

    def getColor(self):
        return self.color

    def setAnimationDelay(self, delay):
        if self.m_timerId != -1:
            self.killTimer(self.m_timerId)

        self.m_delay = delay

        if self.m_timerId != -1:
            self.m_timerId = self.startTimer(self.m_delay)

    def setColor(self, color):
        self.m_color = color
        self.update()

    def timerEvent(self, event):
        self.m_angle = (self.m_angle + 30) % 360
        self.update()

    def paintEvent(self, event):

        width = min(self.width(), self.height())

        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing)

        outerRadius = (width - 1) * 0.5
        innerRadius = (width - 1) * 0.5 * 0.38

        capsuleHeight = outerRadius - innerRadius
        capsuleWidth = capsuleHeight * 0.23 if (width > 32) else capsuleHeight * 0.35
        capsuleRadius = capsuleWidth / 2

        for i in range(0, 12):
            color = QtGui.QColor(self.m_color)

            if self.isAnimated():
                color.setAlphaF(1.0 - (i / 12.0))
            else:
                color.setAlphaF(0.2)

            painter.setPen(QtCore.Qt.NoPen)
            painter.setBrush(color)
            painter.save()
            painter.translate(self.rect().center())
            painter.rotate(self.m_angle - (i * 30.0))
            painter.drawRoundedRect(
                capsuleWidth * -0.5,
                (innerRadius + capsuleHeight) * -1,
                capsuleWidth,
                capsuleHeight,
                capsuleRadius,
                capsuleRadius,
            )
            painter.restore()



# -*- coding: utf-8 -*-
"""
Meta tools for raw datasets
"""
from abc import ABCMeta
from contextlib import suppress


def subclasses(cls):
    """ Return a set of subclasses of ``cls``, including sub-subclasses and so on. """
    direct_subclasses = set(cls.__subclasses__())
    return direct_subclasses.union(
        {s for c in direct_subclasses for s in subclasses(c)}
    )


class MetaRawDataset(ABCMeta):
    """
    Metaclass for AbstractRawDataset.

    This metaclass allows to determine the valid metadata that has been defined using the
    ExperimentalParameter class descriptor as class variables. For example, the AbstractRawDataset
    class already has some built-in ExperimentalParameter descriptors (date, notes, etc.)

    Moreover, all subclasses of types generated by MetaRawDataset have an ``implementations`` attribute
    which points to concrete subclasses, both direct and indirect.
    """

    def __init__(cls, clsname, bases, clsdict):
        super().__init__(clsname, bases, clsdict)

        if not hasattr(cls, "valid_metadata"):
            cls.valid_metadata = set([])

        # valid metadata as defined on the local class
        # Only metadata defined via the ExperimentalParameter descriptor will appear in
        # instance.metadata
        local_valid_metadata = {
            name
            for name, parameter in cls.__dict__.items()
            if isinstance(parameter, ExperimentalParameter)
        }
        cls.valid_metadata = cls.valid_metadata.union(local_valid_metadata)

        # If available, also include valid metadata from superclasses
        with suppress(AttributeError):
            cls.valid_metadata = set.union(cls.valid_metadata, super().valid_metadata)

        # Check if a display_name is available; otherwise, it is the name of the class
        if not hasattr(cls, "display_name"):
            cls.display_name = str(cls.__name__)

    @property
    def implementations(self):
        """ Iterable of concrete implementations. This includes direct subclasses, sub-subclasses, and so on. """
        return subclasses(self)


class MetaHDF5Dataset(ABCMeta):
    """
    Metaclass for DiffractionDataset.

    This metaclass allows to determine the valid metadata that has been defined using the
    HDF5ExperimentalParameter class descriptor as class variables.
    """

    def __init__(cls, clsname, bases, clsdict):
        super().__init__(clsname, bases, clsdict)

        if not hasattr(cls, "valid_metadata"):
            cls.valid_metadata = set([])

        # valid metadata as defined on the local class
        # Only metadata defined via the ExperimentalParameter descriptor will appear in
        # instance.metadata
        local_valid_metadata = {
            name
            for name, parameter in cls.__dict__.items()
            if isinstance(parameter, HDF5ExperimentalParameter)
        }
        cls.valid_metadata = cls.valid_metadata.union(local_valid_metadata)

        # If available, also include valid metadata from superclasses
        with suppress(AttributeError):
            cls.valid_metadata = set.union(cls.valid_metadata, super().valid_metadata)


class ExperimentalParameter:
    """
    Descriptor to experimental parameters and metadata, with default values and
    forced types.

    Parameters
    ----------
    name : str
        Parameter name
    ptype : type or callable
        Parameter type, e.g. float, or callable, e.g. numpy.array.
    default : object or None
        Default value of the parameter. If None, no default value is set. Hence, the
        default value can never be None.
    """

    __slots__ = ("name", "type", "default")

    def __init__(self, name, ptype, default=None):
        self.name = name
        self.type = ptype
        self.default = default

    def __get__(self, instance, cls):
        if instance is None:
            return self
        return instance.__dict__.get(self.name, self.default)

    def __set__(self, instance, value):
        """ If the value cannot be cast to the expected type, a TypeError is raised. """
        try:
            value = self.type(value)
        except ValueError:
            raise TypeError(
                f"Experimental parameter {self.name} expects values of type \
                             {self.type}, but received {value}"
            )
        else:
            instance.__dict__[self.name] = value


class HDF5ExperimentalParameter(ExperimentalParameter):
    """
    Descriptor to experimental parameters and metadata stored in an iris DiffractionDataset.
    These should never be instantiated directly; they are created dynamically at dataset creation.

    Parameters
    ----------
    name : str
        Parameter name
    ptype : type or callable
        Parameter type, e.g. float, or callable, e.g. numpy.array.
    default : object or None
        Default value of the parameter. If None, no default value is set. Hence, the
        default value can never be None.
    """

    def __get__(self, instance, cls):
        value = instance.experimental_parameters_group.attrs.get(
            self.name, default=self.default
        )
        return self.type(value) if value is not None else None

    def __set__(self, instance, value):
        if (value is None) and (self.default is not None):
            value = self.default
        instance.experimental_parameters_group.attrs[self.name] = value

    def __delete__(self, instance):
        del instance.experimental_parameters_group.attrs[self.name]



# -*- coding: utf-8 -*
from glob import glob
import os
import re
from pathlib import Path
from setuptools import setup, find_packages

# To upload to pypi.org:
#   >>> python setup.py sdist
#   >>> twine upload dist/iris-x.x.x.tar.gz

PACKAGE_NAME = "iris-ued"
DESCRIPTION = "Ultrafast electron diffraction data exploration"
URL = "http://iris-ued.readthedocs.io"
DOWNLOAD_URL = "http://github.com/LaurentRDC/iris-ued"
AUTHOR = "Laurent P. René de Cotret"
AUTHOR_EMAIL = "laurent.renedecotret@mail.mcgill.ca"
BASE_PACKAGE = "iris"

base_path = Path(__file__).parent
with open(base_path / BASE_PACKAGE / "__init__.py") as f:
    module_content = f.read()
    VERSION = (
        re.compile(r".*__version__ = \"(.*?)\"", re.S).match(module_content).group(1)
    )
    LICENSE = (
        re.compile(r".*__license__ = \"(.*?)\"", re.S).match(module_content).group(1)
    )

with open("README.md") as f:
    README = f.read()

with open("requirements.txt") as f:
    REQUIREMENTS = [line for line in f.read().split("\n") if len(line.strip())]

exclude = {"exclude": ["docs", "*cache"]}
PACKAGES = [
    BASE_PACKAGE + "." + x
    for x in find_packages(str(base_path / BASE_PACKAGE), **exclude)
]
if BASE_PACKAGE not in PACKAGES:
    PACKAGES.append(BASE_PACKAGE)

if __name__ == "__main__":
    setup(
        name=PACKAGE_NAME,
        description=DESCRIPTION,
        long_description=README,
        long_description_content_type="text/markdown",
        license=LICENSE,
        url=URL,
        download_url=DOWNLOAD_URL,
        version=VERSION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=AUTHOR,
        maintainer_email=AUTHOR_EMAIL,
        install_requires=REQUIREMENTS,
        keywords=["ultrafast electron diffraction visualization pyqtgraph"],
        packages=PACKAGES,
        data_files=[("iris\\gui\\images", glob("iris\\gui\\images\\*.png"))],
        entry_points={"gui_scripts": ["iris = iris.__main__:main"]},
        include_package_data=True,
        project_urls={
            "Documentation": "http://iris-ued.readthedocs.io/en/master/",
            "Source": "https://github.com/LaurentRDC/iris-ued",
            "Tracker": "https://github.com/LaurentRDC/iris-ued/issues",
            "Home": "http://www.physics.mcgill.ca/siwicklab/software.html",
        },
        python_requires=">= 3.7",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Physics",
            "Topic :: Scientific/Engineering :: Chemistry",
            "Topic :: Scientific/Engineering :: Visualization",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
            "Natural Language :: English",
            "Operating System :: OS Independent",
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: Implementation :: CPython",
        ],
    )



# -*- coding: utf-8 -*-
"""
Dialog for azimuthal average of diffraction data
"""
import numpy as np
import pyqtgraph as pg
from PyQt5 import QtCore, QtWidgets
from skued import autocenter
from .controller import WorkThread
from .qbusyindicator import QBusyIndicator
from .symmetrize_dialog import CircleROIWithCenter


normalize_help = """ If checked, all powder patterns will be normalized to their overall intensity.
This can get rid of systematic offsets between patterns at different time-delay. """

explanation = """Drag and resize the red circle until it sits on top of a diffraction ring. 
This allows for easy determination of the picture center. """.replace(
    "\n", ""
)


class AngularAverageDialog(QtWidgets.QDialog):
    """
    Modal dialog to promote a DiffractionDataset to a
    PowderDiffractionDataset
    """

    angular_average_signal = QtCore.pyqtSignal(dict)

    def __init__(self, image, mask, center=None, *args, **kwargs):
        """
        Parameters
        ----------
        image : ndarray
            Diffraction pattern to be displayed.
        """
        super().__init__(*args, **kwargs)
        self.setModal(True)
        self.setWindowTitle("Calculate azimuthal averages")

        # For use with autocenter
        self._image = image
        self._mask = mask

        title = QtWidgets.QLabel("<h2>Azimuthal Average Options<\\h2>")
        title.setTextFormat(QtCore.Qt.RichText)
        title.setAlignment(QtCore.Qt.AlignCenter)

        explanation_label = QtWidgets.QLabel(explanation, parent=self)
        explanation_label.setWordWrap(True)

        self.viewer = pg.ImageView(parent=self)
        self.viewer.setSizePolicy(
            QtWidgets.QSizePolicy.MinimumExpanding,
            QtWidgets.QSizePolicy.MinimumExpanding,
        )
        self.viewer.setImage(image)
        self.center_finder = CircleROIWithCenter(
            pos=np.array(image.shape) / 2 - 100, size=[200, 200], pen=pg.mkPen("r")
        )
        if center is not None and center != (0, 0):
            self.center_finder.set_center(*center)
        self.viewer.getView().addItem(self.center_finder)

        self.partial_circle_btn = QtWidgets.QCheckBox("Restrict azimuthal angle", self)
        self.partial_circle_btn.setChecked(False)

        self.accept_btn = QtWidgets.QPushButton("Calculate", self)
        self.accept_btn.clicked.connect(self.accept)

        self.cancel_btn = QtWidgets.QPushButton("Cancel", self)
        self.cancel_btn.clicked.connect(self.reject)
        self.cancel_btn.setDefault(True)

        self.autocenter_btn = QtWidgets.QPushButton("Autocenter", self)
        self.autocenter_btn.clicked.connect(self.initiate_autocenter)
        self.autocenter_btn.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )
        self.busy_indicator = QBusyIndicator(parent=self)
        autocenter_layout = QtWidgets.QHBoxLayout()
        autocenter_layout.addWidget(self.autocenter_btn)
        autocenter_layout.addWidget(self.busy_indicator)

        self.min_angular_bound_widget = QtWidgets.QDoubleSpinBox(parent=self)
        self.min_angular_bound_widget.setRange(0, 360)
        self.min_angular_bound_widget.setSingleStep(1)
        self.min_angular_bound_widget.setValue(0)
        self.min_angular_bound_widget.setSuffix(" deg")
        self.min_angular_bound_widget.setEnabled(False)
        self.partial_circle_btn.toggled.connect(
            self.min_angular_bound_widget.setEnabled
        )

        self.max_angular_bound_widget = QtWidgets.QDoubleSpinBox(parent=self)
        self.max_angular_bound_widget.setRange(0, 360)
        self.max_angular_bound_widget.setSingleStep
        self.max_angular_bound_widget.setValue(360)
        self.max_angular_bound_widget.setSuffix(" deg")
        self.max_angular_bound_widget.setEnabled(False)
        self.partial_circle_btn.toggled.connect(
            self.max_angular_bound_widget.setEnabled
        )

        self.min_angular_bound_widget.valueChanged.connect(
            self.max_angular_bound_widget.setMinimum
        )
        self.max_angular_bound_widget.valueChanged.connect(
            self.min_angular_bound_widget.setMaximum
        )

        self.normalize_widget = QtWidgets.QCheckBox("Normalize (?)", self)
        self.normalize_widget.setChecked(False)
        self.normalize_widget.setToolTip(normalize_help)

        angle_bounds_layout = QtWidgets.QFormLayout()
        angle_bounds_layout.addRow(self.partial_circle_btn)
        angle_bounds_layout.addRow("Min. angle: ", self.min_angular_bound_widget)
        angle_bounds_layout.addRow("Max. angle: ", self.max_angular_bound_widget)

        btns = QtWidgets.QHBoxLayout()
        btns.addWidget(self.accept_btn)
        btns.addWidget(self.cancel_btn)

        params_layout = QtWidgets.QVBoxLayout()
        params_layout.addWidget(title)
        params_layout.addWidget(explanation_label)
        params_layout.addLayout(autocenter_layout)
        params_layout.addWidget(self.normalize_widget)
        params_layout.addLayout(angle_bounds_layout)
        params_layout.addLayout(btns)

        params_widget = QtWidgets.QFrame(parent=self)
        params_widget.setLayout(params_layout)
        params_widget.setFrameShadow(QtWidgets.QFrame.Sunken)
        params_widget.setFrameShape(QtWidgets.QFrame.Panel)
        params_widget.setSizePolicy(
            QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum
        )

        right_layout = QtWidgets.QVBoxLayout()
        right_layout.addWidget(params_widget)
        right_layout.addStretch()

        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.viewer)
        self.layout.addLayout(right_layout)
        self.setLayout(self.layout)

        self.initiate_autocenter()

    @QtCore.pyqtSlot()
    def accept(self):
        # Calculating the center position assumes that PyQtGraph is configured
        # such that imageAxisOrder == 'row-major'
        center = self.center_finder.center()

        params = {
            "center": center,
            "normalized": self.normalize_widget.isChecked(),
            "angular_bounds": None,
        }  # default

        if self.partial_circle_btn.isChecked():
            params["angular_bounds"] = (
                self.min_angular_bound_widget.value(),
                self.max_angular_bound_widget.value(),
            )

        self.angular_average_signal.emit(params)
        super().accept()

    @QtCore.pyqtSlot()
    def initiate_autocenter(self):
        """Automatically determine the center of an image
        and move the center-finder accordingly"""
        self._worker = AutocenteringThread(
            function=autocenter, kwargs=dict(im=self._image, mask=self._mask)
        )
        self._worker.results_signal.connect(self.set_center)
        self._worker.in_progress_signal.connect(self.busy_indicator.toggle_animation)
        self._worker.in_progress_signal.connect(self.autocenter_btn.setDisabled)
        self._worker.start()

    @QtCore.pyqtSlot(object)
    def set_center(self, rc):
        row, col = rc
        self.center_finder.set_center(col, row)


class AutocenteringThread(WorkThread):
    results_signal = QtCore.pyqtSignal(object)



# -*- coding: utf-8 -*-
"""
Plugin managemenent
===================

Plug-ins are modules implementing a subclass of AbstractRawDataset. 
Plug-ins should be placed in ~\\iris_plugins.

Plug-in classes can be imported from ``iris.plugins``.
"""

from pathlib import Path
from runpy import run_path
from shutil import copy2

PLUGIN_DIR = Path.home() / Path("iris_plugins")
PLUGIN_DIR.mkdir(exist_ok=True)


def load_plugin(path):
    """
    Load an iris plug-in, but do not install it. The plug-in
    will not be available in the next session.

    .. versionadded:: 5.2.0

    Parameters
    ----------
    path : path-like
        Path to the plug-in.
    """
    path = Path(path)
    if not path.exists():
        raise OSError(f"The plug-in file {path} does not exist.")

    try:
        globals().update(run_path(str(path), run_name="iris.plugins"))
    except Exception as e:
        print(f"Failed loading plugin {path} with error: {e}")


def install_plugin(path):
    """
    Install and load an iris plug-in. Installed plug-ins are loaded
    at every iris start-up.

    .. versionadded:: 5.0.4

    Parameters
    ----------
    path : path-like
        Path to the plug-in. This plug-in file will be copied.
    """
    path = Path(path)
    new_path = PLUGIN_DIR / path.name
    copy2(path, new_path)

    load_plugin(new_path)


for fname in PLUGIN_DIR.rglob("*.py"):
    load_plugin(PLUGIN_DIR / fname)
